//++
// eemul.c
//
// Copyright (C) 2000 by Robert Armstrong.
//
//   This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
//   You should have received a copy of the GNU General Public License along
// with this program; if not, visit the website of the Free Software
// Foundation, Inc., www.gnu.org.
//
//
// DESCRIPTION:
//   This file contains a library of interface routines for the EEPROM emulator
// board.  This board emulates two 27256 (32Kx8) EPROM chips arranged in a 32K
// by 16 architecture.  The emulator interfaces to the PC via a standard (well,
// "standard" so long as it supports PS/2 style bidirectional transfers) paral-
// lel port and to the system under test via two 28 pin DIP jumpers.
//
//   The SRAM on the emulator board is not truly dual ported, and so the device
// has two distinct modes - LOAD mode, in which the PC has access to the SRAM,
// and RUN mode in which the target system has access.  The PC controls the
// mode via the parallel port INIT signal.  If the target system is still run-
// ning and tries to read memory while the emulator is in LOAD mode, it will
// always read words of 0xFFFF.
//
//   In LOAD mode, the PC can read or write the emulator's SRAM, and the
// direction of data transfer is controlled by the parallel port SELECT signal.
// The actual data transfer is done via the 8 parallel port data bits, and the
// STROBE signal is used as a data clock.  Writing data from the PC to the
// emulator is fairly conventional, however reading data back from the emulator
// requires that the PC's parallel port support PS/2 style data input trans-
// fers./ Since the emulator's memory is actually 16 bits wide and the port is
// only 8, the AUTOFD signal is used as a HIGH/LOW byte select.
//
//   Finally, the emulator's memory is addressed by a pair of HC590 8 bit
// counters.  These counters are reset on the RUN -> LOAD state transition and
// increment on every read or write cycle.  The PC has no way to directly load
// the current address, and so we're only able to access emulator memory
// sequentially.  Naturally when the emulator is in RUN mode the memory address
// is supplied by the target system and the HC590 address counters are ignored.
//
// REVISION HISTORY:
//   19-DEC-99  RLA     New file.
//--

// Include files...
#include <stdio.h>              // printf(), scanf(), etc....
#include <stdlib.h>             // exit(), etc...
#include <malloc.h>             // malloc(), _fmalloc(), free(), etc...
#include <conio.h>              // _inp() and _outp() ...
#include "romtools.h"           // standard declarations (BYTE, WORD, etc)
#include "eeload.h"             // local declarations for this project

PRIVATE char szSpinner[] = "|/-\\";
#define cbSpinner ((sizeof(szSpinner)-1) / sizeof(char))


// Standard IBM PC parallel port definitions...
#define DATA    0               //  offset of the data port from the base address
#define STATUS  1               //    "     "  "  status  "   "   "    "     "
#define CONTROL 2               //    "     "  "  control "   "   "    "     "
 
//   This bit disables the parallel port output drivers on a PS/2 compatible
// bi-directional port and allows data to be read back from the emulator.  It's
// actually part of the PC's hardware and doesn't have anything directly to do
// with the emulator board.  Note that before data can actually be read the
// emulator's output drivers must be enabled by clearing the WRITE bit in
// addition to setting this bit...
#define READ    0xA0

//   The emulator's DATA CLOCK is STROBE L from the PC.  The parallel port
// hardware inverts STROBE so that setting the bit to zero causes the output
// pin to go high.  Thus from our point of view this bit should normally be
// zero, and then set and cleared to transfer data.  The emulator actually
// latches data on the rising (1 -> 0, remember - it's inverted!) edge of this
// signal.  N.B. the same edge on this bit also increments the HC590 address
// counters on the emulator board.
#define STROBE  0x01

//   The PC's AUTOFD L output drives the LOW/-HIGH emulator signal. Once again,
// the parallel port inverts this line so setting this bit drives LOW/-HIGH
// high which selects the upper byte for reading or writing.  Likewise, clear-
// ing this bit selects the low byte.
#define HIGH    0x02

//  The port's INIT output drives the RUN/-LOAD signal in the emulator, however
// this one isn't inverted so setting the RUN bit drives RUN/-LOAD high and
// puts the emulator in RUN (i.e. emulation) mode.  Before any data can be
// transferred to or from the emulator it must be put in LOAD mode by clearing
// this bit.
#define RUN     0x04

//   Finally, the SELECT L output drives the READ/-WRITE emulator signal and,
// like most of the others, this is inverted by the parallel port hardware.
// Setting this put enables data transfer from the PC to the emulator, and
// clearing it enables the PC to read back data from the emulator.  Note that
// this bit must be coordinated with the READ bit in the port's control
// register so that only one set of drivers is enabled at a time!
#define WRITE   0x08


//++
//   This routine puts the emulator board into a known state, load, and, as a
// side effect, clears the address counters so that the next read or write
// operation starts with SRAM location zero.  There's no way to explicitly load
// the address counters with a specific value, so starting from zero and
// counting up is the only way to get anywhere!
//--
PUBLIC void ResetEmulator (UINT nPort)
{
  //   The HC590 address counters are reset any time the RUN signal is high,
  // and the emulator itself is reset by the RUN -> LOAD transition.  The only
  // way to reset everything is to briefely put it into RUN mode and then
  // return to the LOAD state.  Not accidentally, this also leaves us in the
  // LOAD state, which is where we want to be.
  _outp(nPort+CONTROL, RUN|WRITE);  _outp(nPort+CONTROL, WRITE);

  //   The HC590s used for the address register are a strange beast, and
  // contain both an 8 bit ripple carry counter _and_ eight D flip flops to
  // latch the outputs of the counter.  The chips provide separate clocks for
  // the counters and the latches, but in the emulator board these are tied
  // together.  Setting RUN clears the counter part of the 590, but it does
  // nothing to the latch part, so the actually address inputs to the SRAM
  // haven't changed yet!  We have to clock the latch part to make the address
  // bits go to zero, but this also clocks the counter register and changes it
  // to 1.  The next time we generate a DATA CLOCK, the address latches load
  // the 1 and the counter increments to 2, and so on.  The actually address
  // seen by the SRAMs remains one behind the current count for ever more.
  //
  //   Finally, we want to be very careful not to cause an edge on STROBE while
  // WRITE is asserted - this will _write_ whatever SRAM address happens to be
  // selected now!  It may seem obvious, but it escaped me the first time
  // around...
  _outp(nPort+CONTROL, READ);  _outp(nPort+CONTROL, READ|STROBE);
  _outp(nPort+CONTROL, READ);  _outp(nPort+CONTROL, WRITE);
}


//++
//   This function will put the emulator into the RUN state, which logically
// disconnects it from the PC and allows it to begin emulating memory accesses
// for the system under test...
//--
PUBLIC void RunEmulator (UINT nPort)
{
  _outp(nPort+CONTROL, RUN|WRITE);
}


//++
//   This routine will size the emulator's memory by resetting and then
// repeatedly incrementing the address counters.  The emulator asserts the
// parallel port signal ERROR L whenever the address counter overflows, and we
// can size memory by watching for this.  Note that the emulator's memory is
// implicitly one word (i.e. 16 bits) wide, and this routine returns the size
// in _words_, not bytes!
//--
PUBLIC UINT SizeEmulator (UINT nPort)
{
  UINT nCount = 0;
  ResetEmulator(nPort);  _outp(nPort+CONTROL, READ);
  while ((_inp(nPort+STATUS) & 0x08) != 0) {
    //   If no emulator is connected, then the ERROR L signal never gets
    // asserted no matter how long we wait.  In that case, don't hang up
    // forever!
    if (++nCount == 0) return 0;
    _outp(nPort+CONTROL, READ|STROBE);  _outp(nPort+CONTROL, READ);
  }
  return nCount;
}


//++
//   This routine will download up to 32K words of data to the emulator's SRAM.
// Any amount of data less than 32K may be loaded (e.g. if we're emulating a 2K
// or 8K ROM), but since we've no direct control over the address counters any
// load always starts from SRAM location zero regardless of the size.
//
//   [Actually we could load bytes starting at any offset by first resetting
// the emulator (which sets the address counter to zero), doing read cycles
// until we reach the desired offset, and then switching to writing.  Since
// we've currently no application for this feature, I haven't bothered!]
//
//   Note that the emulator is 16 bits wide, and this routine gives you the
// option of loading either the low byte, the high byte, or both.  For every
// STROBE the emulator will increment the address counter regardles of which
// byte is addressed, so loading both bytes requires two passes.  It's 
// impossible to write (or read) both bytes of the same location in a single
// pass.
//--
PUBLIC void LoadEmulator (UINT nPort, HPBYTE hpbLow, HPBYTE hpbHigh, UINT cbData)
{
  UINT nWord, nSpin;

  //   Reset the emulator (which sets the address counters back to zero), put
  // it in write mode, and load the lower byte of every location.
  if (hpbLow != NULL) {
    ResetEmulator(nPort);  _outp(nPort+CONTROL, WRITE);
    for (nWord = nSpin = 0;  nWord < cbData;  ++nWord) {
      _outp(nPort+DATA, *(hpbLow+nWord));
      _outp(nPort+CONTROL, WRITE|STROBE);  _outp(nPort+CONTROL, WRITE);
      if ((nWord & 0xF) == 0)
        fprintf(stderr, "%c\b", szSpinner[nSpin++ % cbSpinner]);
    }
  }
  
  // Now do it all again, but this time load the high byte.
  if (hpbHigh != NULL) {
    ResetEmulator(nPort);  _outp(nPort+CONTROL, WRITE|HIGH);
    for (nWord = 0;  nWord < cbData;  ++nWord) {
      _outp(nPort+DATA, *(hpbHigh+nWord));
      _outp(nPort+CONTROL, WRITE|HIGH|STROBE);  _outp(nPort+CONTROL, WRITE|HIGH);
      if ((nWord & 0xF) == 0)
        fprintf(stderr, "%c\b", szSpinner[nSpin++ % cbSpinner]);
    }
  }

  // Reset the emulator one last time to leave it in run mode...
  RunEmulator(nPort);
}


//++
//   This routine will upload the emulator's memory into a buffer on the PC.
// It's the opposite of LoadEmulator() and can be used to verify that the data
// was transferred correctly, or as part of a memory diagnostic.  Just like
// loading, since the emulator memory is 16 bits wide every location uploads a
// pair of bytes, and the caller can choose to receive either the low byte, the
// high byte, or both.
//--
PUBLIC void DumpEmulator (UINT nPort, HPBYTE hpbLow, HPBYTE hpbHigh, UINT cbData)
{
  UINT nWord, nSpin;

  //   Reset the emulator and get ready to read data.  Note that this case is
  // more complicated then the WRITE instance because we have to both put the
  // emulator into READ mode (by clearing the WRITE bit) _and_ put the PC's
  // parallel port into INPUT mode (by setting the READ bit).  Naturally this
  // all assumes that this PC's port is capable of PS/2 style bidirectional
  // operation.
  
  // Make the first pass to read the low order bytes...
  if (hpbLow != NULL) {
    ResetEmulator(nPort);  _outp(nPort+CONTROL, READ);
    for (nWord = nSpin = 0;  nWord < cbData;  ++nWord) {
      *(hpbLow+nWord) = _inp(nPort+DATA);
      _outp(nPort+CONTROL, READ|STROBE);  _outp(nPort+CONTROL, READ);
      if ((nWord & 0xF) == 0)
        fprintf(stderr, "%c\b", szSpinner[nSpin++ % cbSpinner]);
    }
  }

  // And then make a second pass to read the high order bytes.
  if (hpbHigh != NULL) {
    ResetEmulator(nPort);  _outp(nPort+CONTROL, READ|HIGH);
    for (nWord = 0;  nWord < cbData;  ++nWord) {
      *(hpbHigh+nWord) |= _inp(nPort+DATA);
      _outp(nPort+CONTROL, READ|HIGH|STROBE);  _outp(nPort+CONTROL, READ|HIGH);
      if ((nWord & 0xF) == 0)
        fprintf(stderr, "%c\b", szSpinner[nSpin++ % cbSpinner]);
    }
  }

  RunEmulator(nPort);
}


//++
//   This function is essentially the same as ReadEmulator(), except that it
// verifies that the data read from the emulator agrees with what is already
// in the buffers.  If any discrepencies are found, an error message is printed
// and FALSE is returned.
//--
PUBLIC BOOL VerifyEmulator (UINT nPort, HPBYTE hpbLow, HPBYTE hpbHigh, UINT cbData)
{
  UINT nWord, nSpin;  BYTE b;  BOOL fError = FALSE;

  // Make the first pass to verify the low order bytes...
  if (hpbLow != NULL) {
    ResetEmulator(nPort);  _outp(nPort+CONTROL, READ);
    for (nWord = nSpin = 0;  nWord < cbData;  ++nWord) {
      b = _inp(nPort+DATA);
      if (*(hpbLow+nWord) != b) {
        fprintf(stderr, "EELOAD: verification error (low byte): address 0x%04X, expected 0x%02X, read 0x%02X", nWord, *(hpbLow+nWord), b);
        fError = TRUE;
      }
      _outp(nPort+CONTROL, READ|STROBE);  _outp(nPort+CONTROL, READ);
      if ((nWord & 0xF) == 0)
        fprintf(stderr, "%c\b", szSpinner[nSpin++ % cbSpinner]);
    }
  }

  // And verify the high bytes...
  if (hpbHigh != NULL) {
    ResetEmulator(nPort);  _outp(nPort+CONTROL, READ|HIGH);
    for (nWord = 0;  nWord < cbData;  ++nWord) {
      b = _inp(nPort+DATA);
      if (*(hpbHigh+nWord) != b) {
        fprintf(stderr, "EELOAD: verification error (high byte): address 0x%04X, expected 0x%02X, read 0x%02X", nWord, *(hpbLow+nWord), b);
        fError = TRUE;
      }
      _outp(nPort+CONTROL, READ|HIGH|STROBE);  _outp(nPort+CONTROL, READ|HIGH);
      if ((nWord & 0xF) == 0)
        fprintf(stderr, "%c\b", szSpinner[nSpin++ % cbSpinner]);
    }
  }

  RunEmulator(nPort);
  return !fError;
}
